using System.Text.RegularExpressions;
using System.Xml;
using System.Xml.Linq;
using System.Xml.Schema;
using FitConnect.Exceptions;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using Newtonsoft.Json.Schema;

namespace FitConnect.Validation;

public class SchemaValidationVisitor {
    private readonly string? _schemaStoragePath;
    private readonly ILogger? _logger;

    public SchemaValidationVisitor(string schemaStoragePath, ILogger? logger = null) {
        _schemaStoragePath = schemaStoragePath;
        _logger = logger;
    }

    /// <summary>
    /// Checks if the data is valid against the schema
    /// </summary>
    /// <param name="data">Data to test against the schema</param>
    /// <param name="schema">Can be the schema itself or the path to a schema file</param>
    /// <remarks>XML is not ready for production, yet</remarks>
    public async Task<bool> ValidateSubmission(string data, string schema) {
        if (string.IsNullOrWhiteSpace(data)) {
            throw new Exception("Data is required");
        }


        if (string.IsNullOrWhiteSpace(schema)) {
            throw new Exception("Schema is required");
        }

        var (schemaContent, extension) = await GetSchemaContent(schema);

        switch (extension) {
            case "json":
                return ValidationJsonData(data, schemaContent);
            case "xml":
                _logger?.LogWarning("XML is not ready for production, yet");
                return ValidateXmlData(data, schemaContent);
            default:
                throw new Exception("Invalid schema");
        }
    }

    internal async Task<(string schema, string schemaType)> GetSchemaContent(string schema) {
        var validWinPathRegex = new Regex(@"^(?:[\w]\:|\\)(\\[A-Za-z_\-\s0-9\.]+)+\.(json|xml)$");
        var validLinuxPathRegex = new Regex(@"^(\.)?(/[^/ ]*)+\.(json|xml)$");
        var validUrlRegex = new Regex(@"^http(s)?://([\w-]+.)+[\w-]+(/[\w- ./?%&=])?$");

        if (validWinPathRegex.IsMatch(schema) || validLinuxPathRegex.IsMatch(schema)) {
            var extension = Path.GetExtension(schema);
            if (string.IsNullOrWhiteSpace(extension)) {
                throw new Exception("Invalid schema");
            }

            var schemaContent = await File.ReadAllTextAsync(schema);
            return (schemaContent, extension.Replace(".", ""));
        }

        if (validUrlRegex.IsMatch(schema)) {
            string? content = null;
            string? extension = null;
            try {
                content = await new HttpClient().GetStringAsync(schema);
                extension = Path.GetExtension(schema).Replace(".", "");
            }
            catch (Exception e) {
                _logger?.LogError(e, "Schema download failed");
            }

            if (_schemaStoragePath != null &&
                !string.IsNullOrWhiteSpace(extension) &&
                !string.IsNullOrWhiteSpace(content)) {
                var fileName = Path.GetFileName(schema);
                var filePath = Path.Combine(_schemaStoragePath, fileName);
                if (!Directory.Exists(_schemaStoragePath))
                    Directory.CreateDirectory(_schemaStoragePath);
                await File.WriteAllTextAsync(filePath, content);
            }
            else if (_schemaStoragePath != null) {
                _logger?.LogWarning("Schema download failed, trying to use local schema");
                var fileName = Path.GetFileName(schema);
                var filePath = Path.Combine(_schemaStoragePath, fileName);
                if (File.Exists(filePath)) {
                    content = await File.ReadAllTextAsync(filePath);
                    extension = Path.GetExtension(schema).Replace(".", "");
                }
                else {
                    _logger?.LogError("Local schema not found");
                    throw new FitConnectSubscriberException("Data validation failed: Missing Schema");
                }
            }

            if (string.IsNullOrWhiteSpace(content) || string.IsNullOrWhiteSpace(extension)) {
                _logger?.LogError("Schema storage path is not set and download failed");
                throw new FitConnectSubscriberException(
                    "Data validation failed, Schema storage path is not set and download failed");
            }


            return (content, extension);
        }

        if (schema.IsJsonSchema())
            return (schema, "json");
        if (schema.IsXmlSchema())
            return (schema, "xml");

        throw new Exception("Invalid schema");
    }


    private bool ValidationJsonData(string data, string schema) {
        var schemaJson = JSchema.Parse(schema);
        var dataJson = JObject.Parse(data);
        var isValid = dataJson.IsValid(schemaJson, out IList<string> errors);
        if (!isValid) {
            _logger?.LogError("Invalid data: {ValidationErrors}", string.Join(", ", errors));
            return false;
        }

        return isValid;
    }

    private bool ValidateXmlData(string data, string schema) {
        var schemaSet = new XmlSchemaSet();
        schemaSet.Add("", XmlReader.Create(new StringReader(schema)));
        var doc = XDocument.Parse(data);
        doc.Validate(schemaSet,
            (o, e) => { _logger?.LogError("Invalid data: {ValidationError}", e.Message); });
        return true;
    }
}

public static class SchemaCheck {
    public static bool IsJsonSchema(this string input) {
        try {
            JSchema.Parse(input);
            return true;
        }
        catch (JsonReaderException) {
            return false;
        }
        catch (JsonSchemaException) {
            return false;
        }
    }


    public static bool IsXmlSchema(this string input) {
        try {
            var schemaSet = new XmlSchemaSet();
            schemaSet.Add("", XmlReader.Create(new StringReader(input)));

            schemaSet.Compile();

            return true;
        }
        catch (XmlException) {
            return false;
        }
        catch (XmlSchemaException) {
            return false;
        }
    }
}
